import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:http_server/http_server.dart';
import 'package:logging/logging.dart';
import 'package:path/path.dart' as pathlib;

const UDP_PORT = 10003, _NAME = "network_file";
final _bcastAddr = InternetAddress("255.255.255.255"), _l = Logger(_NAME);

class FileIndex {
  final Directory root;
  final VirtualDirectory vDir;

  FileIndex(Directory root)
      : this.root = root.absolute,
        vDir = VirtualDirectory(root.path);

  File getFile(String key) {
    return File(root.path + "/" + key);
  }

  void serveFile(String key, HttpRequest request) {
    vDir.serveFile(getFile(key), request);
  }
}

Uint8List _dumpInt(int value) {
  return Uint64List.fromList([value]).buffer.asUint8List();
}

List _loadInt(List value) {
  return [
    Uint8List.fromList(value.sublist(0, 8)).buffer.asUint64List()[0],
    value.sublist(8),
  ];
}

typedef bool ShouldSAcceptDiscovery(FileSystemEntity file, String extra);

class FileDiscoveryServer {
  final FileIndex index;
  final int transferServerPort, uid;
  final ShouldSAcceptDiscovery shouldAcceptDiscovery;

  RawDatagramSocket sock;
  StreamSubscription<RawSocketEvent> listener;

  FileDiscoveryServer(
    this.index,
    this.transferServerPort,
    this.uid, {
    this.shouldAcceptDiscovery,
  });

  Future<void> bind() async {
    sock = await RawDatagramSocket.bind(InternetAddress.anyIPv4, UDP_PORT);
  }

  Future<void> handleRequest(Datagram req) async {
    final data = req.data;

    if (data.isEmpty || data[0] == 0) {
      _l.fine("peer discovery request { address: ${req.address} }");
      sock.send([0], req.address, req.port);
      return;
    }

    final parts = _loadInt(data);
    if (parts[0] == uid) return;

    final params = jsonDecode(utf8.decode(parts[1])),
        path = params[0],
        extra = params[1],
        file = index.getFile(path),
        exists = await file.exists();

    _l.fine(
      "file discovery request { path: $path, extra: $extra file: $file exists: $exists, address: ${req.address} }",
    );

    if (!exists || !(shouldAcceptDiscovery?.call(file, extra) ?? true)) return;
    sock.send(_dumpInt(transferServerPort), req.address, req.port);
  }

  void serveForever() {
    _l.info(
      "FileDiscoveryServer listenng on: udp://${sock.address.address}:${sock.port}",
    );
    listener = sock.listen((event) {
      final req = sock.receive();
      if (req == null) return;
      try {
        handleRequest(req);
      } catch (e, trace) {
        _l.fine("error in discovery server", e, trace);
      }
    });
  }

  Future<void> close() async {
    sock?.close();
    sock = null;
    await listener?.cancel();
    listener = null;
    _l.info("FileDiscoveryServer closed!");
  }
}

class FileTransferServer {
  final FileIndex index;

  HttpServer server;
  StreamSubscription<HttpRequest> listener;

  FileTransferServer(this.index);

  Future<void> bind() async {
    server = await HttpServer.bind(InternetAddress.anyIPv4, 0);
  }

  void serveForever() {
    _l.info(
      "FileTransferServer listenng on: tcp://${server.address.address}:${server.port}",
    );
    listener = server.listen((request) {
      String key = pathlib.relative(request.uri.path, from: "/");
      _l.fine(
        "transfer request { key: $key ${request.connectionInfo.remoteAddress} }",
      );
      index.serveFile(key, request);
    });
  }

  Future<void> close() async {
    await server?.close();
    server = null;
    await listener?.cancel();
    listener = null;
    _l.info("FileTransferServer closed!");
  }
}

class NetworkFile {
  FileDiscoveryServer discoveryServer;
  FileTransferServer transferServer;

  Future<void> start(
    Directory root, {
    ShouldSAcceptDiscovery shouldAcceptDiscovery,
  }) async {
    var index = FileIndex(root);

    transferServer = FileTransferServer(index);
    await transferServer.bind();
    transferServer.serveForever();

    discoveryServer = FileDiscoveryServer(
      index,
      transferServer.server.port,
      hashCode,
      shouldAcceptDiscovery: shouldAcceptDiscovery,
    );
    await discoveryServer.bind();
    discoveryServer.serveForever();
  }

  Future<void> stop() async {
    await discoveryServer?.close();
    discoveryServer = null;
    await transferServer?.close();
    transferServer = null;
  }

  Future<String> findPeer({
    Duration timeout: const Duration(seconds: 30),
  }) async {
    final sock = await RawDatagramSocket.bind(InternetAddress.anyIPv4, 0);
    sock.broadcastEnabled = true;

    sock.send([0], _bcastAddr, UDP_PORT);

    final completer = Completer<String>();
    StreamSubscription sub;

    sub = sock.timeout(timeout).listen((event) {
      final data = sock.receive(), response = data?.data;
      if (response == null) return;

      final addr = data.address.address;
      _l.info("found peer at: $addr");

      sub.cancel();
      completer.complete(addr);
    }, onError: (error, stackTrace) {
      sub.cancel();
      completer.completeError(error, stackTrace);
    });

    return await completer.future;
  }

  Future<Uri> findFile(
    String path, {
    Duration timeout: const Duration(seconds: 30),
    String extra,
  }) async {
    final sock = await RawDatagramSocket.bind(InternetAddress.anyIPv4, 0);
    sock.broadcastEnabled = true;

    final payload = _dumpInt(hashCode) + utf8.encode(jsonEncode([path, extra]));
    sock.send(payload, _bcastAddr, UDP_PORT);

    final completer = Completer<Uri>();
    StreamSubscription sub;

    sub = sock.timeout(timeout).listen((event) {
      final data = sock.receive(), response = data?.data;
      if (response == null) return;

      final uri = Uri(
        scheme: "http",
        host: data.address.address,
        port: _loadInt(response)[0],
        path: path,
      );
      _l.info("found file at url: $uri");

      sub.cancel();
      completer.complete(uri);
    }, onError: (error, stackTrace) {
      sub.cancel();
      completer.completeError(error, stackTrace);
    });

    return await completer.future;
  }

  NetworkFile._internal();

  static NetworkFile _instance;

  factory NetworkFile.getInstance() {
    if (_instance == null) _instance = NetworkFile._internal();
    return _instance;
  }
}
